#include <QtGui/QMouseEvent>
#include <cassert>
#include <cmath>
#include <GL/glu.h>
#include "GLWidget.h"

const double ALPHA0 = (-10.);
const double BETA0 = 18.;

static const GLfloat EARTH_RADIUS = 0.3;

static const GLfloat MOON_RADIUS  = 0.1;
static const GLfloat MOON_ORBIT_RADIUS  = 1.2;
static const GLfloat MOON_ROTATION_SPEED = 6.;  // Orbit speed degrees/sec
static const GLfloat MOON_SPIN_SPEED = 6.;      // Rotation speed
static const GLfloat EARTH_SPIN_SPEED = 18;     // 28.*MOON_ROTATION_SPEED;

static const int ANIMATION_DT = 60;     // In milliseconds

static const double COORD_SCALE = 1.5;

GLWidget::GLWidget(QWidget *parent):
    QGLWidget(parent),
    xmin(-COORD_SCALE * 4./3.),
    xmax(COORD_SCALE *4./3.),
    ymin(-COORD_SCALE),
    ymax(COORD_SCALE),
    alpha(ALPHA0),
    beta(BETA0),
    mouseX(-1),     // Undefined
    mouseY(-1),     // Undefined
    timer(0),
    animationTime(),
    earthSpin(0.),
    moonAlpha(0.),
    moonSpin(0.),
    quadric(0)
{
    setMouseTracking(true);

    animationTime = QTime::currentTime();
    timer = new QTimer(this);
    connect(timer, SIGNAL(timeout()), this, SLOT(animate()));
    timer->start(30);
}

void GLWidget::initializeGL() {
    glEnable(GL_DEPTH_TEST);
    glEnable(GL_LIGHTING);
    glDisable(GL_CULL_FACE);
    glEnable(GL_NORMALIZE);
    glDepthFunc(GL_LEQUAL);

    // Lighting
    glEnable(GL_LIGHT0);
    glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, 1);

    GLfloat pos[4];
    pos[0] = 0.; pos[1] = 0.; pos[2] = 1.; pos[3] = 0.; // Directional light
    glLightfv(GL_LIGHT0, GL_POSITION, pos);

    GLfloat light[4];
    light[0] = 0.25; light[1] = 0.25; light[2] = 0.25; light[3] = 1.;
    glLightfv(GL_LIGHT0, GL_AMBIENT, light);

    light[0] = 1.; light[1] = 1.; light[2] = 1.; light[3] = 1.;
    glLightfv(GL_LIGHT0, GL_DIFFUSE, light);

    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
}

void GLWidget::resizeGL(int w, int h) {
    glViewport(0, 0, w, h);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();

    if (w == 0)
        w = 1;
    if (h == 0)
        h = 1;
    double aspect = (double) w / (double) h;
    assert(aspect > 0.);

    ymin = xmin / aspect;
    ymax = (-ymin);
    double depth = ymax - ymin;
    glOrtho(
        xmin, xmax,
        ymin, ymax,
        -10. * depth, 10. * depth
    );
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
}

void GLWidget::paintGL() {
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    // Rotate the complete model
    glRotatef(beta, 1., 0., 0.);
    glRotatef(alpha, 0., 1., 0.);

    // Light of Sun
    glEnable(GL_LIGHT0);        // Turn on the first light source

    // Position of light source
    GLfloat pos[4];
    pos[0] = 10.; pos[1] = 0.; pos[2] = 10.;
    pos[3] = 0.;                // Directional light
    glLightfv(GL_LIGHT0, GL_POSITION, pos);

    // Color of Sun
    GLfloat color[4];
    color[0] = 0.25; color[1] = 0.25; color[2] = 0.25; color[3] = 1.;
    glLightfv(GL_LIGHT0, GL_AMBIENT, color);
    color[0] = 1.; color[1] = 1.; color[2] = 0.8;
    glLightfv(GL_LIGHT0, GL_DIFFUSE, color);
    color[2] = 1.;
    glLightfv(GL_LIGHT0, GL_SPECULAR, color);

    render();           // Draw all 3D objects
}

void GLWidget::mousePressEvent(QMouseEvent * /* event */) {
    mouseX = (-1);      // Undefined
    mouseY = (-1);      // Undefined
}

void GLWidget::mouseReleaseEvent(QMouseEvent * /* event */) {
    mouseX = (-1);      // Undefined
    mouseY = (-1);      // Undefined
}

void GLWidget::mouseMoveEvent(QMouseEvent *event) {
    //... printf("%d, %d\n", event->x(), event->y());
    int x = event->x();
    int y = event->y();
    if ((event->buttons() & Qt::LeftButton) != 0) {
        if (mouseX >= 0) {
            int dx = x - mouseX;
            int dy = y - mouseY;
            if (beta > 100. || beta < (-100.))
                dx = -dx;

            alpha += dx * 0.1;
            if (fabs(alpha) > 360.)
                alpha -= ((int) alpha / 360) * 360.;

            beta += dy * 0.1;
            if (fabs(beta) > 360.)
                beta -= ((int) beta / 360) * 360.;

            repaint();
        }
        mouseX = x; mouseY = y;

    } else {
        mouseX = (-1);      // Undefined
        mouseY = (-1);      // Undefined
    }
}

void GLWidget::keyPressEvent(QKeyEvent* event) {
    switch (event->key()) {
    case Qt::Key_Escape:
    case Qt::Key_Q:
        close();
        break;
    case Qt::Key_Space:
        alpha = ALPHA0;
        beta = BETA0;
        repaint();
        break;
    default:
        event->ignore();
        break;
    }
}

//
// Draw a 3D model
//
void GLWidget::render() {
    GLfloat color[4];

    glClearColor(0.2, 0.3, 0.5, 1.); // Background color: dark blue
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    // Draw Earth
    if (quadric == 0) {
        quadric = gluNewQuadric();    // Create a Quadric object
        gluQuadricNormals(quadric, GLU_SMOOTH);
    }
    color[0] = 0.3; color[1] = 0.8; color[2] = 0.6; color[3] = 1.;
    glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, color);
    glMaterialf(GL_FRONT, GL_SHININESS, 0.3);

    glRotatef(earthSpin, 0., 1., 0.);
    glRotatef(90., 1., 0., 0.);
    gluSphere(
        quadric,
        EARTH_RADIUS,
        18,     // Num. slices (similar to lines of longitude)
        10      // Num. stacks (similar to lines of latitude)
    );
    // Draw meridians / parallels
    color[0] = 0.2; color[1] = 0.5; color[2] = 0.4; color[3] = 1.;
    glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, color);
    gluQuadricDrawStyle(quadric, GLU_LINE); // Draw lines only
    gluSphere(
        quadric,
        EARTH_RADIUS + 0.005,
        18,     // Num. slices (similar to lines of longitude)
        10      // Num. stacks (similar to lines of latitude)
    );
    gluQuadricDrawStyle(quadric, GLU_FILL); // Restore the normal draw style
    glRotatef(-90., 1., 0., 0.);
    glRotatef(-earthSpin, 0., 1., 0.);

    // Draw Moon
    glRotatef(moonAlpha, 0., 1., 0.); // Orbit rotation

    // Shift Moon center
    glTranslatef(MOON_ORBIT_RADIUS, 0., 0.);
    color[0] = 0.2; color[1] = 0.6; color[2] = 1.;
    glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, color);
    glMaterialf(GL_FRONT, GL_SHININESS, 0.6);

    glRotatef(-90., 1., 0., 0.);
    glRotatef(moonSpin, 0., 1., 0.);  // Moon spin

    gluSphere(
        quadric,
        MOON_RADIUS,
        16,     // Num. slices (similar to lines of longitude)
        8       // Num. stacks (similar to lines of latitude)
    );
}

void GLWidget::animate() {
    QTime curtime = QTime::currentTime();
    if (curtime < animationTime)
        return;

    int cur_msec = curtime.msec();
    int prev_msec = animationTime.msec();
    if (cur_msec < prev_msec)
        cur_msec += 1000;
    // Time interval from the previous redraw
    int dt_msec = ANIMATION_DT + (cur_msec - prev_msec);
    GLfloat dt = (GLfloat) dt_msec / 1000.;     // Convert to seconds

    moonAlpha += MOON_ROTATION_SPEED * dt;
    if (moonAlpha > 360.)
        moonAlpha -= 360.;
    moonSpin += MOON_SPIN_SPEED * dt;
    if (moonSpin > 360.)
        moonSpin -= 360.;
    earthSpin += EARTH_SPIN_SPEED * dt;
    if (earthSpin > 360.)
        earthSpin -= 360.;

    repaint();

    animationTime = curtime.addMSecs(ANIMATION_DT);
}
